@@ -8,6 +8,7 @@ import os
 import pipes
 import re
 import shutil
+import struct
 import subprocess as sp
 import sys
 import tempfile
@@ -24,10 +25,10 @@ from debughelpers import (Error, Info, UnloggedError, UnloggedInfo,
 						  UnloggedWarning, Warn)
 from log import LogFile, safe_print
 from meta import name as appname
-from multiprocess import mp
+from multiprocess import mp, pool_slice
 from options import debug, verbose
 from util_os import getenvu, quote_args, which
-from util_str import safe_basestring, safe_str, safe_unicode
+from util_str import make_filename_safe, safe_basestring, safe_str, safe_unicode
 import CGATS
 import colormath
 import config
@@ -44,7 +45,8 @@ def _mp_xicclu(chunk, thread_abort_event, progress_queue, profile_filename,
 			   pcs=None, scale=1, cwd=None, startupinfo=None, use_icclu=False,
 			   use_cam_clipping=False, logfile=None,
 			   show_actual_if_clipped=False, input_encoding=None,
-			   output_encoding=None, abortmessage="Aborted"):
+			   output_encoding=None, abortmessage="Aborted", output_format=None,
+			   reverse=False):
 	if not config.cfg.items(config.ConfigParser.DEFAULTSECT):
 		config.initcfg()
 	profile = ICCP.ICCProfile(profile_filename)
@@ -57,7 +59,7 @@ def _mp_xicclu(chunk, thread_abort_event, progress_queue, profile_filename,
 	num_subchunks = 50
 	subchunksize = float(len(chunk)) / num_subchunks
 	for i in xrange(num_subchunks):
-		if thread_abort_event.is_set():
+		if thread_abort_event is not None and thread_abort_event.is_set():
 			xicclu.exit()
 			return Info(abortmessage)
 		end = int(math.ceil(subchunksize * (i + 1)))
@@ -68,7 +70,7 @@ def _mp_xicclu(chunk, thread_abort_event, progress_queue, profile_filename,
 			progress_queue.put(perc - prevperc)
 			prevperc = perc
 	xicclu.exit()
-	return xicclu.get()
+	return xicclu.get(output_format=output_format, reverse=reverse)
 
 
 def _mp_generate_B2A_clut(chunk, thread_abort_event, progress_queue,
@@ -462,12 +464,23 @@ class Xicclu(WorkerBase):
 			else:
 				profile = ICCP.ICCProfile(profile)
 		is_profile = isinstance(profile, ICCP.ICCProfile)
+		if is_profile:
+			if profile.fileName:
+				prefix = os.path.basename(profile.fileName)
+			else:
+				prefix = make_filename_safe(profile.getDescription(),
+											concat=False) + profile_ext
+			prefix += "-"
+			if (profile.version >= 4 and
+				not profile.convert_iccv4_tags_to_iccv2()):
+				raise Error("\n".join([lang.getstr("profile.iccv4.unsupported"),
+												   profile.getDescription()]))
 		if not profile.fileName or not os.path.isfile(profile.fileName):
 			if not cwd:
 				cwd = self.create_tempdir()
 				if isinstance(cwd, Exception):
 					raise cwd
-			fd, profile.fileName = tempfile.mkstemp(profile_ext, dir=cwd)
+			fd, profile.fileName = tempfile.mkstemp("", prefix, dir=cwd)
 			stream = os.fdopen(fd, "wb")
 			profile.write(stream)
 			stream.close()
@@ -486,11 +499,13 @@ class Xicclu(WorkerBase):
 		xicclu = safe_str(xicclu)
 		cwd = safe_str(cwd)
 		args = [xicclu, "-s%s" % scale]
+		self.show_actual_if_clipped = False
 		if utilname == "xicclu":
 			if (is_profile and
 				show_actual_if_clipped and "A2B0" in profile.tags and
 				("B2A0" in profile.tags or direction == "if")):
 				args.append("-a")
+				self.show_actual_if_clipped = True
 			if use_cam_clipping:
 				args.append("-b")
 			if get_argyll_version("xicclu") >= [1, 6]:
@@ -651,7 +666,7 @@ class Xicclu(WorkerBase):
 							   tuple(safe_unicode(s) for s in 
 									 (self.tempdir, exception)))
 	
-	def get(self, raw=False, get_clip=False):
+	def get(self, raw=False, get_clip=False, output_format=None, reverse=False):
 		if raw:
 			if self.sessionlogfile:
 				self.sessionlogfile.write("\n".join(self.output))
@@ -662,7 +677,12 @@ class Xicclu(WorkerBase):
 		for i, line in enumerate(self.output):
 			line = line.strip()
 			if line.startswith("["):
-				if self.sessionlogfile:
+				if parsed and get_clip and self.show_actual_if_clipped:
+					parts = line.strip("[]").split(",")
+					actual = [float(v) for v in parts[0].split()[1:4]]  # Actual CIE
+					actual.append(float(parts[1].split()[-1]))  # deltaE
+					parsed[-1].append(actual)
+				elif self.sessionlogfile:
 					self.sessionlogfile.write(line)
 				continue
 			elif not "->" in line:
@@ -675,8 +695,13 @@ class Xicclu(WorkerBase):
 			clip = parts.pop() == "(clip)"
 			if clip:
 				parts.pop()
-			parsed.append([float(n) / self.output_scale for n in parts])
-			if get_clip:
+			out = [float(n) / self.output_scale for n in parts]
+			if output_format:
+				if reverse:
+					out = out[::-1]
+				out = "".join(struct.pack(output_format[0], round(v * output_format[1])) for v in out)
+			parsed.append(out)
+			if get_clip and not self.show_actual_if_clipped:
 				parsed[-1].append(clip)
 			j += 1
 		if self.sessionlogfile:
@@ -700,9 +725,68 @@ class Xicclu(WorkerBase):
 		def fget(self):
 			if self.worker:
 				return self.worker.thread_abort
-			return False
+			return None
 		
 		def fset(self, v):
 			pass
 		
 		return locals()
+
+
+class MP_Xicclu(Xicclu):
+
+	def __init__(self, profile, intent="r", direction="f", order="n",
+				 pcs=None, scale=1, cwd=None, startupinfo=None, use_icclu=False,
+				 use_cam_clipping=False, logfile=None, worker=None,
+				 show_actual_if_clipped=False, input_encoding=None,
+				 output_encoding=None, output_format=None, reverse=False,
+				 output_stream=None):
+		WorkerBase.__init__(self)
+		self.logfile = logfile
+		self.worker = worker
+		self.output_stream = output_stream
+		self._in = []
+		self._args = (profile.fileName, intent, direction, order,
+					  pcs, scale, cwd, startupinfo, use_icclu,
+					  use_cam_clipping, None,
+					  show_actual_if_clipped, input_encoding,
+					  output_encoding, lang.getstr("aborted"), output_format,
+					  reverse)
+		self._out = []
+		num_cpus = mp.cpu_count()
+		if isinstance(profile.tags.get("A2B0"), ICCP.LUT16Type):
+			size = profile.tags.A2B0.clut_grid_steps
+			self.num_workers = min(max(num_cpus, 1), size)
+			if num_cpus > 2:
+				self.num_workers = int(self.num_workers * 0.75)
+			self.num_batches = size // 9
+		else:
+			if num_cpus > 2:
+				self.num_workers = 2
+			else:
+				self.num_workers = num_cpus
+			self.num_batches = 1
+
+	def __call__(self, idata):
+		self._in.append(idata)
+
+	def close(self):
+		pass
+
+	def exit(self):
+		pass
+
+	def spawn(self):
+		pass
+
+	def get(self):
+
+		for slices in pool_slice(_mp_xicclu, self._in, self._args, {},
+								 self.num_workers, self.thread_abort,
+								 self.logfile, num_batches=self.num_batches):
+			if self.output_stream:
+				for row in slices:
+					self.output_stream.write(row)
+			else:
+				self._out.extend(slices)
+		return self._out
